import re
import pandas as pd
from ._describer import desc_pipeline, desc_node
from .adapter  import get_adapter
class PipelineGroup:
    def __init__(
        self, name, role, processor=None, edges=None, method=None, parent=None, adapter=None, params=None
    ):
        self.name = name
        self.role = role  # 'stage' or 'head'
        self.processor = processor
        self.edges = edges if edges is not None else {}
        self.method = method
        self.parent = parent  # parent group name (str)
        self.adapter = adapter
        self.params = params if params is not None else {}
        self.children = []  # child group names
        self.nodes = []  # node names in this group
        self.attrs = None

    def get_attrs(self, grps):
        if self.attrs is not None:
            return self.attrs
        if self.parent is None:
            parent_attrs = {
                'edges': {},
                'params': {},
                'processor': None,
                'method': None,
                'adapter': None
            }
        else:
            parent_attrs = grps[self.parent].get_attrs(grps)
        edges = self.edges.copy()
        if parent_attrs['edges'] is not None:
            for k, v in parent_attrs['edges'].items():
                edges[k] = edges.get(k, []) + v
        params = self.params.copy()
        if parent_attrs['params'] is not None:
            for k, v in parent_attrs['params'].items():
                if k not in params:
                    params[k] = v
        processor = parent_attrs['processor'] if self.processor is None else self.processor
        if self.adapter is None:
            if parent_attrs['adapter'] is not None:
                adapter = parent_attrs['adapter']
            else:
                adapter = None
        else:
            adapter = self.adapter
        self.attrs = {
            'name': self.name,
            'edges': edges,
            'parent': self.parent,
            'adapter': adapter,
            'params': params,
            'children': self.children,
        }
        for i in ['role', 'processor', 'method']:
            self.attrs[i] = parent_attrs.get(i) if getattr(self, i) is None else getattr(self, i)

        return self.attrs

    def update_attrs(self):
        self.attrs = None

    def copy(self):
        ret = PipelineGroup(
            self.name, self.role, self.processor, self.edges.copy(),
            self.method, self.parent, self.adapter, self.params.copy()
        )
        ret.children = self.children.copy()
        ret.nodes = self.nodes.copy()
        return ret

class PipelineNode:
    def __init__(
        self, name, grp, processor=None, edges=None, method=None, adapter=None, params=None
    ):
        self.name = name
        self.grp = grp  # group name (str)
        self.processor = processor
        self.edges = edges if edges is not None else {}
        self.method = method
        self.adapter = adapter
        self.params = params if params is not None else {}

        self.output_edges = []  # 이 노드를 입력으로 사용하는 노드들의 이름
        self.attrs = None

    def copy(self):
        ret = PipelineNode(
            self.name, self.grp, self.processor, self.edges.copy(),
            self.method, self.adapter, self.params.copy()
        )
        ret.output_edges = self.output_edges.copy()
        return ret

    def get_attrs(self, grps):
        if self.attrs is not None:
            return self.attrs
        grp_attrs = grps[self.grp].get_attrs(grps)
        edges = self.edges.copy()
        if grp_attrs['edges'] is not None:
            for k, v in grp_attrs['edges'].items():
                edges[k] = edges.get(k, []) + v
        params = self.params.copy()
        if grp_attrs['params'] is not None:
            for k, v in grp_attrs['params'].items():
                if k not in params:
                    params[k] = v
        processor = grp_attrs['processor'] if self.processor is None else self.processor
        if self.adapter is None:
            if grp_attrs['adapter'] is None:
                adapter = get_adapter(processor)
            else:
                adapter = grp_attrs['adapter']
        else:
            adapter = self.adapter
        self.attrs = {
            'name': self.name,
            'grp': self.grp,
            'edges': edges,
            'processor': processor,
            'adapter': adapter,
            'params': params,
            'method': grp_attrs.get('method') if self.method is None else self.method,
        }

        return self.attrs

    def update_attrs(self):
        self.attrs = None


class Pipeline:
    def __init__(self):
        self.nodes = {}
        self.grps = {}
        self.nodes = {None: PipelineNode("Data_Source", None, None, None, None, None)}

    def copy(self):
        ret = Pipeline()
        ret.grps = {k: v.copy() for k, v in self.grps.items()}
        ret.nodes = {k: v.copy() for k, v in self.nodes.items()}
        return ret

    def copy_stage(self):
        ret = Pipeline()

        stage_grp_names = {name for name, grp in self.grps.items() if grp.role == 'stage'}

        for name in stage_grp_names:
            grp = self.grps[name].copy()
            grp.children = [c for c in grp.children if c in stage_grp_names]
            if grp.parent not in stage_grp_names:
                grp.parent = None
            ret.grps[name] = grp

        stage_node_names = set()
        for name, node in self.nodes.items():
            if name is None:
                continue
            if node.grp in stage_grp_names:
                stage_node_names.add(name)

        for name in stage_node_names:
            node = self.nodes[name].copy()
            node.output_edges = [e for e in node.output_edges if e in stage_node_names]
            ret.nodes[name] = node

        data_source = self.nodes[None].copy()
        data_source.output_edges = [e for e in data_source.output_edges if e in stage_node_names]
        ret.nodes[None] = data_source

        return ret

    def copy_nodes(self, node_names):
        needed_nodes = set()
        queue = list(node_names)

        while queue:
            name = queue.pop(0)
            if name is None or name in needed_nodes:
                continue
            if name not in self.nodes:
                continue
            needed_nodes.add(name)
            attrs = self.nodes[name].get_attrs(self.grps)
            for edge_list in attrs.get('edges', {}).values():
                for edge_name, _ in edge_list:
                    if edge_name is not None and edge_name not in needed_nodes:
                        queue.append(edge_name)

        needed_grps = set()
        for name in needed_nodes:
            grp_name = self.nodes[name].grp
            while grp_name is not None and grp_name not in needed_grps:
                needed_grps.add(grp_name)
                grp_name = self.grps[grp_name].parent

        ret = Pipeline()

        for name in needed_grps:
            grp = self.grps[name].copy()
            grp.children = [c for c in grp.children if c in needed_grps]
            grp.nodes = [n for n in grp.nodes if n in needed_nodes]
            if grp.parent not in needed_grps:
                grp.parent = None
            ret.grps[name] = grp

        for name in needed_nodes:
            node = self.nodes[name].copy()
            node.output_edges = [e for e in node.output_edges if e in needed_nodes]
            ret.nodes[name] = node

        data_source = self.nodes[None].copy()
        data_source.output_edges = [e for e in data_source.output_edges if e in needed_nodes]
        ret.nodes[None] = data_source

        return ret

    def _validate_name(self, name):
        if name is None:
            return

        if '__' in name:
            raise ValueError(f"Name '{name}' cannot contain '__'")

        invalid_chars = ['/', '\\', '\0', '<', '>', ':', '"', '|', '?', '*']
        for char in invalid_chars:
            if char in name:
                raise ValueError(f"Name '{name}' cannot contain '{char}'")

    def _find_descendants(self, node_name):
        descendants = set()
        queue = [node_name]

        while queue:
            current = queue.pop(0)

            if current not in self.nodes:
                continue

            for child_name in self.nodes[current].output_edges:
                if child_name not in descendants:
                    descendants.add(child_name)
                    queue.append(child_name)

        return descendants

    def _check_cycle(self, node_name, new_edges):
        descendants = self._find_descendants(node_name)

        cycle_edges = []
        for key, edge_list in new_edges.items():
            for edge_name, _ in edge_list:
                if edge_name is None:
                    continue

                if edge_name not in self.nodes:
                    continue

                if edge_name in descendants:
                    cycle_edges.append(edge_name)

        if cycle_edges:
            return True, cycle_edges
        return False, []

    def _check_edges(self, edges):
        if edges is None or len(edges) == 0:
            return False
        for key, edge_list in edges.items():
            for name, _ in edge_list:
                if name is None:
                    continue
                if name not in self.nodes:
                    raise ValueError(f"Edge node '{name}' not found")
                node_grp = self.nodes[name].grp
                if self.grps[node_grp].role != 'stage':
                    raise ValueError(f"Edge node '{name}' must be a stage node, got '{self.grps[node_grp].role}'")
        return True

    def _get_all_nodes_in_grp(self, grp):
        result = list(grp.nodes)
        for child_name in grp.children:
            child_grp = self.grps[child_name]
            result.extend(self._get_all_nodes_in_grp(child_grp))
        return result

    def _get_affected_nodes(self, nodes):
        priorities = {}
        queue = []

        for node_name in nodes:
            priorities[node_name] = 1
            queue.append((node_name, 1))

        while queue:
            current_node, current_priority = queue.pop(0)

            descendants = self._find_descendants(current_node)

            for desc_node in descendants:
                new_priority = current_priority + 1
                if desc_node not in priorities or priorities[desc_node] < new_priority:
                    priorities[desc_node] = new_priority
                    queue.append((desc_node, new_priority))

        sorted_nodes = sorted(priorities.items(), key=lambda x: x[1])
        return [i[0] for i in sorted_nodes if i[0] is not None]

    def set_grp(
            self, name, role=None, processor=None, edges=None, method=None, parent=None, adapter=None, params=None, exist='skip'
        ):
        self._validate_name(name)
        if name in self.nodes:
            raise ValueError(f"Name '{name}' already exists as a node")
        if edges is None:
            edges = {}

        if parent is not None:
            if parent not in self.grps:
                raise ValueError(f"Parent group '{parent}' not found")
            if role is None:
                role = self.grps[parent].role
        if role not in ['stage', 'head']:
            raise ValueError(f"Role must be 'stage' or 'head', got '{role}'")

        if name not in self.grps:
            self._check_edges(edges)
            grp = PipelineGroup(
                name, role, processor=processor, edges=edges, method=method, parent=parent, adapter=adapter, params=params
            )

            if parent is not None:
                self.grps[parent].children.append(name)

            self.grps[name] = grp
            return {
                "result": "new", "grp": grp, "affected_nodes": list()
            }
        elif exist == 'skip':
            grp = self.grps[name]
            return {"result": "skip", "grp": grp, "affected_nodes": list()}
        elif exist == 'error':
            raise ValueError(f"Group '{name}' already exists.")

        old_grp = self.grps[name]
        if old_grp.role != role:
            raise ValueError(f"Cannot change role of group '{name}': existing '{old_grp.role}', requested '{role}'")
        grp = old_grp.copy()

        parent_changed = False
        old_parent = old_grp.parent

        if old_parent != parent:
            parent_changed = True
            if old_parent is not None:
                self.grps[old_parent].children.remove(name)
            grp.parent = parent
            if parent is not None:
                self.grps[parent].children.append(name)

        if processor is not None:
            grp.processor = processor
        if edges is not None and len(edges) > 0:
            grp.edges = edges
        if method is not None:
            grp.method = method
        if adapter is not None:
            grp.adapter = adapter
        if params is not None:
            grp.params.update(params)

        grp.update_attrs()
        attrs = grp.get_attrs(self.grps)
        new_edges = attrs['edges']
        if len(new_edges) > 0:
            affected_nodes = self._get_all_nodes_in_grp(grp)

            for node_name in affected_nodes:
                if node_name not in self.nodes:
                    continue

                node = self.nodes[node_name]
                node.update_attrs()
                node_attrs = node.get_attrs(self.grps)
                node_own_edges = node_attrs.get('edges', {})

                final_edges = {k: list(v) for k, v in new_edges.items()}
                for k, v in node_own_edges.items():
                    if k in final_edges:
                        final_edges[k].extend(v)
                    else:
                        final_edges[k] = list(v)

                has_cycle, cycle_edges = self._check_cycle(node_name, final_edges)
                if has_cycle:
                    cycle_info = ", ".join([f"'{e}'" for e in cycle_edges])
                    raise ValueError(f"Cannot update group '{name}': node '{node_name}' would create cycle through edge(s) {cycle_info}")
        else:
            affected_nodes = list()

        self.grps[name] = grp
        return {
            "result": "update", "affected_nodes": affected_nodes, "old_grp": old_grp, "grp": grp
        }
    
    def get_grp(self, name):
        return self.grps.get(name, None)
    
    def rename_grp(self, name_from, name_to):
        self._validate_name(name_to)

        if name_from not in self.grps:
            raise ValueError(f"Group '{name_from}' not found")
        if name_to in self.grps:
            raise ValueError(f"Group '{name_to}' already exists")

        old_grp = self.grps[name_from]
        grp = old_grp.copy()
        grp.name = name_to
        if grp.parent is not None:
            self.grps[grp.parent].children.remove(name_from)
            self.grps[grp.parent].children.append(name_to)

        for node_name in grp.nodes:
            self.nodes[node_name].grp = name_to
            self.nodes[node_name].update_attrs()

        for child_name in grp.children:
            self.grps[child_name].parent = name_to
            self.grps[child_name].update_attrs()

        del self.grps[name_from]
        self.grps[name_to] = grp

    def remove_grp(self, name):
        if name not in self.grps:
            raise ValueError(f"Group '{name}' not found")

        grp = self.grps[name]

        if len(grp.children) > 0:
            raise ValueError(f"Cannot remove group '{name}': has {len(grp.children)} child group(s)")

        if len(grp.nodes) > 0:
            raise ValueError(f"Cannot remove group '{name}': has {len(grp.nodes)} node(s)")

        if grp.parent is not None:
            self.grps[grp.parent].children.remove(name)

        del self.grps[name]

    def get_parents(self, node_name):
        if node_name not in self.nodes:
            return []

        node = self.nodes[node_name]
        if node.grp is None:
            return []

        result = []
        current_grp = self.grps.get(node.grp)

        while current_grp is not None:
            result.append(current_grp.name)
            current_grp = self.grps.get(current_grp.parent) if current_grp.parent else None

        return result

    def get_node_names(self, query):
        if query is None:
            node_names = list(self.nodes.keys())
        elif isinstance(query, list):
            node_names = [n for n in query if n in self.nodes]
        elif isinstance(query, str):
            pat = re.compile(query)
            node_names = [k for k in self.nodes.keys() if k is not None and pat.search(k)]
        else:
            raise ValueError(f"query must be None, list, or str, got {type(query)}")
        return node_names

    def remove_node(self, name):
        if name not in self.nodes:
            raise ValueError(f"Node '{name}' not found")

        if name is None:
            raise ValueError("Cannot remove DataSource node")

        descendants = self._find_descendants(name)
        if descendants:
            descendants_list = sorted(descendants)
            raise ValueError(f"Cannot remove node '{name}': has {len(descendants)} dependent node(s): {descendants_list}")

        node = self.nodes[name]
        node_attr = node.get_attrs(self.grps)
        self._update_output_edges(name, node_attr['edges'], None)

        grp_name = node.grp
        if grp_name is not None and grp_name in self.grps:
            grp = self.grps[grp_name]
            if name in grp.nodes:
                grp.nodes.remove(name)

        del self.nodes[name]

    def _update_output_edges(self, node_name, old_edges, new_edges):
        if old_edges is not None:
            for key, edge_list in old_edges.items():
                for edge_name, _ in edge_list:
                    if edge_name in self.nodes:
                        parent_node = self.nodes[edge_name]
                        if node_name in parent_node.output_edges:
                            parent_node.output_edges.remove(node_name)

        if new_edges is not None:
            for key, edge_list in new_edges.items():
                for edge_name, _ in edge_list:
                    if edge_name in self.nodes:
                        parent_node = self.nodes[edge_name]
                        if node_name not in parent_node.output_edges:
                            parent_node.output_edges.append(node_name)

    def set_node(
        self, name, grp, processor=None, edges=None, method=None, adapter=None, params=None, exist='skip'
    ):
        self._validate_name(name)

        if name in self.grps:
            raise ValueError(f"Name '{name}' already exists as a group")

        if grp not in self.grps:
            raise ValueError(f"Group '{grp}' not found")

        if edges is None:
            edges = {}
        if params is None:
            params = {}

        self._check_edges(edges)

        is_update = name in self.nodes
        if is_update:
            if exist == 'skip':
                return {'result': 'skip', 'affected_nodes': [], 'old_obj': self.nodes[name], 'obj': self.nodes[name]}
            elif exist == 'error':
                raise ValueError(f"Node '{name}' already exists.")

        old_edges = None
        old_output_edges = None
        old_node = None
        if is_update:
            old_node = self.nodes[name]
            old_edges = old_node.get_attrs(self.grps)['edges']
            old_output_edges = old_node.output_edges

        node = PipelineNode(
            name, grp, processor, edges, method=method, adapter=adapter, params=params
        )

        grp_obj = self.grps[grp]
        attrs = node.get_attrs(self.grps)

        if attrs.get('processor') is None:
            raise ValueError(f"Cannot create node '{name}': processor is required")

        if attrs.get('method') is None:
            raise ValueError(f"Cannot create node '{name}': method is required")

        if len(attrs.get('edges', {})) == 0:
            raise ValueError(f"Cannot create node '{name}': edges is required")

        has_cycle, cycle_edges = self._check_cycle(name, attrs['edges'])
        if has_cycle:
            cycle_info = ", ".join([f"'{e}'" for e in cycle_edges])
            raise ValueError(f"Cannot add node '{name}': would create cycle through edge(s) {cycle_info}")

        self._update_output_edges(name, old_edges, attrs['edges'])

        if old_output_edges is not None:
            node.output_edges = old_output_edges

        if name not in grp_obj.nodes:
            grp_obj.nodes.append(name)

        if is_update:
            affected_nodes = list(self._find_descendants(name))
            old_grp_name = old_node.grp
            if old_grp_name != grp and old_grp_name in self.grps:
                old_grp = self.grps[old_grp_name]
                if name in old_grp.nodes:
                    old_grp.nodes.remove(name)
        else:
            affected_nodes = list()

        self.nodes[name] = node

        return {
            'result': 'update' if is_update else 'new',
            'affected_nodes': affected_nodes,
            'old_obj': old_node,
            'obj': node
        }

    def get_node(self, name):
        return self.nodes.get(name, None)

    def get_node_attrs(self, name):
        node = self.get_node(name)
        return node.get_attrs(self.grps)

    def desc_pipeline(self, max_depth=None, direction='TD'):
        """파이프라인 구조를 Mermaid Markdown으로 반환

        Args:
            max_depth: 최대 표시 깊이 (None이면 무제한)
            direction: 그래프 방향 ('TD': Top-Down, 'LR': Left-Right)
        """
        return desc_pipeline(self, max_depth, direction)

    def compare_nodes(self, nodes):
        attrs_map = {n: self.get_node_attrs(n) for n in nodes}

        groups = {}
        for name in nodes:
            proc = attrs_map[name]['processor']
            proc_name = proc.__name__ if proc is not None else 'None'
            groups.setdefault(proc_name, []).append(name)

        result = {}
        for proc_name, group_nodes in groups.items():
            rows = {name: {} for name in group_nodes}

            # params
            all_param_keys = sorted({k for n in group_nodes for k in attrs_map[n]['params']})
            for name in group_nodes:
                params = attrs_map[name]['params']
                for k in all_param_keys:
                    rows[name][('params', k)] = params.get(k, None)

            # edges (X only) - stage node별 변수 비교
            stage_vars = {}
            for name in group_nodes:
                x_entries = attrs_map[name]['edges'].get('X', [])
                for sn, var_spec in x_entries:
                    if sn not in stage_vars:
                        stage_vars[sn] = {}
                    if name not in stage_vars[sn]:
                        stage_vars[sn][name] = []
                    if var_spec is None:
                        stage_vars[sn][name].append(None)
                    elif isinstance(var_spec, (list, tuple)):
                        stage_vars[sn][name].extend(var_spec)
                    else:
                        stage_vars[sn][name].append(var_spec)

            for sn, node_vars in stage_vars.items():
                sn_str = str(sn) if sn is not None else 'DataSource'
                for name in group_nodes:
                    if name not in node_vars:
                        node_vars[name] = []

                repr_map = {}
                var_sets = {}
                for name in group_nodes:
                    s = set()
                    for v in node_vars[name]:
                        r = repr(v)
                        s.add(r)
                        repr_map[r] = v
                    var_sets[name] = s

                if len({frozenset(s) for s in var_sets.values()}) <= 1:
                    continue

                non_empty = [s for s in var_sets.values() if s]
                common_reprs = set.intersection(*non_empty) if non_empty else set()
                common_vars = sorted([repr_map[r] for r in common_reprs], key=repr)
                col_2 = f"{sn_str} [{', '.join(str(v) for v in common_vars)}]" if common_vars else sn_str

                for name in group_nodes:
                    diff_reprs = var_sets[name] - common_reprs
                    diff_vars = sorted([repr_map[r] for r in diff_reprs], key=repr)
                    rows[name][('X', col_2)] = diff_vars if diff_vars else []

            df = pd.DataFrame.from_dict(rows, orient='index')
            if len(df.columns) > 0:
                df.columns = pd.MultiIndex.from_tuples(df.columns)
                diff_cols = [c for c in df.columns if len({repr(v) for v in df[c]}) > 1]
                df = df[diff_cols]
            result[proc_name] = df

        return result

    def desc_node(self, node_name, direction='TD', show_params=False):
        """특정 노드까지의 연결 구조를 Mermaid Markdown으로 반환

        Args:
            node_name: 대상 노드 이름
            direction: 그래프 방향 ('TD': Top-Down, 'LR': Left-Right)
            show_params: True이면 노드의 파라미터 정보를 표시 (default: False)
        """
        return desc_node(self, node_name, direction, show_params)


import pytest
import numpy as np
import pandas as pd

from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection import ShuffleSplit, KFold

from mllabs._experimenter import Experimenter
from mllabs._inferencer import Inferencer
from mllabs._data_wrapper import unwrap


@pytest.fixture
def sample_data():
    np.random.seed(42)
    n = 100
    return pd.DataFrame({
        'f1': np.random.randn(n),
        'f2': np.random.randn(n),
        'f3': np.random.randn(n),
        'target': np.random.randint(0, 2, n),
    })


@pytest.fixture
def exp(tmp_path, sample_data):
    e = Experimenter(
        data=sample_data,
        path=tmp_path / 'exp',
        sp=ShuffleSplit(n_splits=2, test_size=0.2, random_state=42),
        sp_v=KFold(n_splits=3, shuffle=True, random_state=42),
    )
    e.set_grp('scale', role='stage', processor=StandardScaler,
              method='transform', edges={'X': [(None, ['f1', 'f2', 'f3'])]})
    e.set_node('scaler', grp='scale')
    e.set_grp('model', role='head', processor=DecisionTreeClassifier,
              method='predict',
              edges={'X': [('scaler', None)], 'y': [(None, 'target')]},
              params={'max_depth': 3, 'random_state': 42})
    e.set_node('dt', grp='model')
    e.build()
    e.exp()
    return e


@pytest.fixture
def trained_trainer(exp):
    trainer = exp.add_trainer('t1')
    trainer.select_head(['dt'])
    trainer.train()
    return trainer


class TestToInferencer:
    def test_basic_creation(self, trained_trainer):
        inf = trained_trainer.to_inferencer()
        assert isinstance(inf, Inferencer)
        assert inf.n_splits == trained_trainer.get_n_splits()
        assert inf.selected_stages == trained_trainer.selected_stages
        assert inf.selected_heads == trained_trainer.selected_heads

    def test_node_objs_are_processor_lists(self, trained_trainer):
        inf = trained_trainer.to_inferencer()
        for name, objs in inf.node_objs.items():
            assert isinstance(objs, list)
            assert len(objs) == inf.n_splits
            assert hasattr(objs[0], 'process')

    def test_minimal_pipeline(self, trained_trainer):
        inf = trained_trainer.to_inferencer()
        assert 'scaler' in inf.pipeline.nodes
        assert 'dt' in inf.pipeline.nodes

    def test_not_trained_raises(self, exp):
        trainer = exp.add_trainer('t_no_train')
        trainer.select_head(['dt'])
        with pytest.raises(RuntimeError, match="not built"):
            trainer.to_inferencer()

    def test_v_stored(self, trained_trainer):
        inf = trained_trainer.to_inferencer(v=[0])
        assert inf.v == [0]


class TestProcess:
    def test_mean_agg(self, trained_trainer, sample_data):
        inf = trained_trainer.to_inferencer()
        result = inf.process(sample_data, agg='mean')
        assert result.get_shape()[0] == len(sample_data)

    def test_mode_agg(self, trained_trainer, sample_data):
        inf = trained_trainer.to_inferencer()
        result = inf.process(sample_data, agg='mode')
        assert result.get_shape()[0] == len(sample_data)

    def test_callable_agg(self, trained_trainer, sample_data):
        inf = trained_trainer.to_inferencer()
        result = inf.process(sample_data, agg=lambda results: results[0])
        assert result.get_shape()[0] == len(sample_data)

    def test_none_agg(self, trained_trainer, sample_data):
        inf = trained_trainer.to_inferencer()
        results = inf.process(sample_data, agg=None)
        assert isinstance(results, list)
        assert len(results) == inf.n_splits

    def test_v_parameter(self, exp, sample_data):
        exp.set_grp('model_proba', role='head', processor=DecisionTreeClassifier,
                    method='predict_proba',
                    edges={'X': [('scaler', None)], 'y': [(None, 'target')]},
                    params={'max_depth': 3, 'random_state': 42})
        exp.set_node('dt_proba', grp='model_proba')
        exp.build()
        exp.exp()
        trainer = exp.add_trainer('t_proba')
        trainer.select_head(['dt_proba'])
        trainer.train()
        inf = trainer.to_inferencer(v=slice(-1, None))
        result = inf.process(sample_data)
        assert result.get_shape()[1] == 1

    def test_single_split(self, exp, sample_data):
        trainer = exp.add_trainer('t_nosplit', splitter=None)
        trainer.select_head(['dt'])
        trainer.train()
        inf = trainer.to_inferencer()
        result = inf.process(sample_data)
        assert result.get_shape()[0] == len(sample_data)

    def test_unknown_agg_raises(self, trained_trainer, sample_data):
        inf = trained_trainer.to_inferencer()
        with pytest.raises(ValueError, match="Unknown agg"):
            inf.process(sample_data, agg='unknown')


class TestSaveLoad:
    def test_save_load_roundtrip(self, trained_trainer, tmp_path):
        inf = trained_trainer.to_inferencer()
        save_path = tmp_path / 'inferencer'
        inf.save(save_path)

        loaded = Inferencer.load(save_path)
        assert loaded.n_splits == inf.n_splits
        assert loaded.selected_stages == inf.selected_stages
        assert loaded.selected_heads == inf.selected_heads
        assert set(loaded.node_objs.keys()) == set(inf.node_objs.keys())

    def test_loaded_process_matches(self, trained_trainer, sample_data, tmp_path):
        inf = trained_trainer.to_inferencer()
        save_path = tmp_path / 'inferencer'
        inf.save(save_path)

        loaded = Inferencer.load(save_path)
        original = inf.process(sample_data, agg=None)
        loaded_result = loaded.process(sample_data, agg=None)

        assert len(original) == len(loaded_result)
        for orig, load in zip(original, loaded_result):
            np.testing.assert_array_equal(unwrap(orig), unwrap(load))

    def test_save_creates_file(self, trained_trainer, tmp_path):
        inf = trained_trainer.to_inferencer()
        save_path = tmp_path / 'inferencer'
        inf.save(save_path)
        assert (save_path / '__inferencer.pkl').exists()

    def test_save_load_with_v(self, trained_trainer, sample_data, tmp_path):
        exp = trained_trainer
        inf = trained_trainer.to_inferencer(v=[0])
        save_path = tmp_path / 'inferencer_v'
        inf.save(save_path)

        loaded = Inferencer.load(save_path)
        assert loaded.v == [0]



import pytest
import numpy as np
from mllabs._data_wrapper import (
    DataWrapper, PandasWrapper, NumpyWrapper, wrap, unwrap
)

pd = pytest.importorskip("pandas", reason="pandas not installed")

try:
    import polars as pl
    HAS_POLARS = True
except ImportError:
    HAS_POLARS = False

try:
    import cudf
    HAS_CUDF = True
except ImportError:
    HAS_CUDF = False

skip_polars = pytest.mark.skipif(not HAS_POLARS, reason="polars not installed")
skip_cudf = pytest.mark.skipif(not HAS_CUDF, reason="cudf not installed")


# ── fixtures ──

@pytest.fixture
def pdf():
    return pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [7, 8, 9]})

@pytest.fixture
def pdf1():
    return pd.DataFrame({'x': [10, 20, 30]})

@pytest.fixture
def narr():
    return np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

@pytest.fixture
def narr1():
    return np.array([[10], [20], [30]])

@pytest.fixture
def pldf():
    if not HAS_POLARS:
        return None
    return pl.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [7, 8, 9]})

@pytest.fixture
def pldf1():
    if not HAS_POLARS:
        return None
    return pl.DataFrame({'x': [10, 20, 30]})


# ── wrap / unwrap ──

class TestWrapUnwrap:
    def test_wrap_pandas(self, pdf):
        w = wrap(pdf)
        assert isinstance(w, PandasWrapper)

    def test_wrap_numpy(self, narr):
        w = wrap(narr)
        assert isinstance(w, NumpyWrapper)

    @skip_polars
    def test_wrap_polars(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = wrap(pldf)
        assert isinstance(w, PolarsWrapper)

    def test_wrap_none(self):
        assert wrap(None) is None

    def test_unwrap_none(self):
        assert unwrap(None) is None

    def test_unwrap_native(self, pdf):
        assert unwrap(pdf) is pdf

    def test_unwrap_wrapper(self, pdf):
        w = PandasWrapper(pdf)
        assert unwrap(w) is pdf

    def test_wrap_unsupported(self):
        with pytest.raises(TypeError):
            wrap("not a data object")


# ── PandasWrapper ──

class TestPandasWrapper:
    def test_iloc_list(self, pdf):
        w = PandasWrapper(pdf)
        result = w.iloc([0, 2])
        assert isinstance(result, PandasWrapper)
        assert result.get_shape() == (2, 3)
        assert result.to_native().iloc[0]['a'] == 1

    def test_iloc_slice(self, pdf):
        w = PandasWrapper(pdf)
        result = w.iloc(slice(1, 3))
        assert result.get_shape() == (2, 3)

    def test_select_columns_single(self, pdf):
        w = PandasWrapper(pdf)
        result = w.select_columns('a')
        assert isinstance(result.to_native(), pd.Series)

    def test_select_columns_list(self, pdf):
        w = PandasWrapper(pdf)
        result = w.select_columns(['a', 'c'])
        assert result.get_columns() == ['a', 'c']

    def test_get_columns(self, pdf):
        w = PandasWrapper(pdf)
        assert w.get_columns() == ['a', 'b', 'c']

    def test_get_columns_series(self):
        s = pd.Series([1, 2, 3], name='x')
        w = PandasWrapper(s)
        assert w.get_columns() == ['x']

    def test_get_shape(self, pdf):
        w = PandasWrapper(pdf)
        assert w.get_shape() == (3, 3)

    def test_get_index(self, pdf):
        w = PandasWrapper(pdf)
        idx = w.get_index()
        assert list(idx) == [0, 1, 2]

    def test_concat_axis1(self, pdf, pdf1):
        w1 = PandasWrapper(pdf)
        w2 = PandasWrapper(pdf1)
        result = PandasWrapper.concat([w1, w2], axis=1)
        assert result.get_shape() == (3, 4)
        assert 'x' in result.get_columns()

    def test_concat_axis0(self, pdf):
        w1 = PandasWrapper(pdf)
        w2 = PandasWrapper(pdf.copy())
        result = PandasWrapper.concat([w1, w2], axis=0)
        assert result.get_shape() == (6, 3)

    def test_to_native(self, pdf):
        w = PandasWrapper(pdf)
        assert w.to_native() is pdf

    def test_to_array(self, pdf):
        w = PandasWrapper(pdf)
        arr = w.to_array()
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (3, 3)

    def test_getitem_str(self, pdf):
        w = PandasWrapper(pdf)
        result = w['a']
        assert isinstance(result, PandasWrapper)

    def test_getitem_list(self, pdf):
        w = PandasWrapper(pdf)
        result = w[['a', 'b']]
        assert result.get_columns() == ['a', 'b']

    # from_output
    def test_from_output_none(self):
        assert PandasWrapper.from_output(None) is None

    def test_from_output_dataframe(self, pdf):
        result = PandasWrapper.from_output(pdf)
        assert isinstance(result, PandasWrapper)
        assert result.get_columns() == ['a', 'b', 'c']

    def test_from_output_dataframe_rename(self, pdf):
        result = PandasWrapper.from_output(pdf, column_names=['x', 'y', 'z'])
        assert result.get_columns() == ['x', 'y', 'z']

    def test_from_output_ndarray_2d(self, narr):
        result = PandasWrapper.from_output(narr, column_names=['a', 'b', 'c'])
        assert isinstance(result, PandasWrapper)
        assert result.get_columns() == ['a', 'b', 'c']

    def test_from_output_ndarray_1d(self):
        arr = np.array([1, 2, 3])
        result = PandasWrapper.from_output(arr)
        assert result.get_shape() == (3, 1)

    def test_from_output_unsupported(self):
        with pytest.raises(TypeError):
            PandasWrapper.from_output("bad")

    # squeeze
    def test_squeeze_single_col(self, pdf1):
        w = PandasWrapper(pdf1)
        result = w.squeeze()
        assert isinstance(result, PandasWrapper)
        native = result.to_native()
        assert isinstance(native, pd.Series)
        assert native.name == 'x'

    def test_squeeze_multi_col(self, pdf):
        w = PandasWrapper(pdf)
        result = w.squeeze()
        native = result.to_native()
        assert isinstance(native, pd.DataFrame)
        assert native.shape == (3, 3)

    def test_squeeze_series(self):
        s = pd.Series([1, 2, 3], name='y')
        w = PandasWrapper(s)
        result = w.squeeze()
        assert isinstance(result.to_native(), pd.Series)

    # aggregation
    def test_simple(self, pdf):
        w = PandasWrapper(pdf)
        result = DataWrapper.simple(iter([w]))
        assert result is w

    def test_mean(self, pdf):
        w1 = PandasWrapper(pdf.copy())
        w2 = PandasWrapper(pdf.copy() * 3)
        result = DataWrapper.mean(iter([w1, w2]))
        native = unwrap(result)
        assert native['a'].tolist() == [2.0, 4.0, 6.0]

    def test_mode(self):
        df1 = pd.DataFrame({'a': [1, 2, 3]})
        df2 = pd.DataFrame({'a': [1, 2, 4]})
        df3 = pd.DataFrame({'a': [1, 5, 4]})
        result = PandasWrapper.mode(iter([wrap(df1), wrap(df2), wrap(df3)]))
        native = unwrap(result)
        assert native['a'].iloc[0] == 1


# ── NumpyWrapper ──

class TestNumpyWrapper:
    def test_iloc_list(self, narr):
        w = NumpyWrapper(narr)
        result = w.iloc([0, 2])
        assert isinstance(result, NumpyWrapper)
        assert result.get_shape() == (2, 3)

    def test_iloc_slice(self, narr):
        w = NumpyWrapper(narr)
        result = w.iloc(slice(0, 2))
        assert result.get_shape() == (2, 3)

    def test_select_columns_single(self, narr):
        w = NumpyWrapper(narr, columns=['a', 'b', 'c'])
        result = w.select_columns('b')
        np.testing.assert_array_equal(result.to_native(), np.array([2, 5, 8]))

    def test_select_columns_multi(self, narr):
        w = NumpyWrapper(narr, columns=['a', 'b', 'c'])
        result = w.select_columns(['a', 'c'])
        assert result.get_shape() == (3, 2)
        assert result.get_columns() == ['a', 'c']

    def test_select_columns_not_found(self, narr):
        w = NumpyWrapper(narr, columns=['a', 'b', 'c'])
        with pytest.raises(KeyError):
            w.select_columns('z')

    def test_get_columns_default(self, narr):
        w = NumpyWrapper(narr)
        assert w.get_columns() == [0, 1, 2]

    def test_get_columns_1d(self):
        w = NumpyWrapper(np.array([1, 2, 3]))
        assert w.get_columns() == [0]

    def test_get_shape(self, narr):
        w = NumpyWrapper(narr)
        assert w.get_shape() == (3, 3)

    def test_get_index(self, narr):
        w = NumpyWrapper(narr)
        assert list(w.get_index()) == [0, 1, 2]

    def test_concat_axis1(self, narr):
        w1 = NumpyWrapper(narr, columns=['a', 'b', 'c'])
        w2 = NumpyWrapper(np.array([[10], [20], [30]]), columns=['d'])
        result = NumpyWrapper.concat([w1, w2], axis=1)
        assert result.get_shape() == (3, 4)
        assert result.get_columns() == ['a', 'b', 'c', 'd']

    def test_concat_axis0(self, narr):
        w1 = NumpyWrapper(narr)
        w2 = NumpyWrapper(narr.copy())
        result = NumpyWrapper.concat([w1, w2], axis=0)
        assert result.get_shape() == (6, 3)

    def test_to_native(self, narr):
        w = NumpyWrapper(narr)
        assert w.to_native() is narr

    def test_to_array(self, narr):
        w = NumpyWrapper(narr)
        assert w.to_array() is narr

    # from_output
    def test_from_output_none(self):
        assert NumpyWrapper.from_output(None) is None

    def test_from_output_ndarray(self, narr):
        result = NumpyWrapper.from_output(narr)
        assert isinstance(result, NumpyWrapper)
        assert result.get_columns() == [0, 1, 2]

    def test_from_output_1d(self):
        arr = np.array([1, 2, 3])
        result = NumpyWrapper.from_output(arr)
        assert result.get_columns() == [0]

    def test_from_output_dataframe(self, pdf):
        result = NumpyWrapper.from_output(pdf)
        assert isinstance(result, NumpyWrapper)
        assert result.get_columns() == ['a', 'b', 'c']

    def test_from_output_unsupported(self):
        with pytest.raises(TypeError):
            NumpyWrapper.from_output("bad")

    # squeeze
    def test_squeeze_2d_single_col(self, narr1):
        w = NumpyWrapper(narr1)
        result = w.squeeze()
        assert isinstance(result, NumpyWrapper)
        assert result.to_native().ndim == 1
        np.testing.assert_array_equal(result.to_native(), [10, 20, 30])

    def test_squeeze_2d_multi_col(self, narr):
        w = NumpyWrapper(narr)
        result = w.squeeze()
        assert result.to_native().shape == (3, 3)

    def test_squeeze_1d(self):
        arr = np.array([1, 2, 3])
        w = NumpyWrapper(arr)
        result = w.squeeze()
        assert result.to_native().ndim == 1

    # aggregation
    def test_mean(self, narr):
        w1 = NumpyWrapper(narr.copy())
        w2 = NumpyWrapper(narr.copy() * 3)
        result = DataWrapper.mean(iter([w1, w2]))
        native = unwrap(result)
        np.testing.assert_array_equal(native[0], [2.0, 4.0, 6.0])

    def test_mode(self):
        a1 = np.array([1, 2, 3])
        a2 = np.array([1, 2, 4])
        a3 = np.array([1, 5, 4])
        result = NumpyWrapper.mode(iter([wrap(a1), wrap(a2), wrap(a3)]))
        native = unwrap(result)
        assert native[0] == 1
        assert native[2] == 4


# ── PolarsWrapper ──

@skip_polars
class TestPolarsWrapper:
    def test_iloc_list(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf)
        result = w.iloc([0, 2])
        assert isinstance(result, PolarsWrapper)
        assert result.get_shape() == (2, 3)

    def test_iloc_slice(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf)
        result = w.iloc(slice(1, 3))
        assert result.get_shape() == (2, 3)

    def test_select_columns(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf)
        result = w.select_columns(['a', 'c'])
        assert result.get_columns() == ['a', 'c']

    def test_get_columns(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf)
        assert w.get_columns() == ['a', 'b', 'c']

    def test_get_shape(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf)
        assert w.get_shape() == (3, 3)

    def test_get_index(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf)
        assert list(w.get_index()) == [0, 1, 2]

    def test_concat_axis1(self, pldf, pldf1):
        from mllabs._data_wrapper import PolarsWrapper
        w1 = PolarsWrapper(pldf)
        w2 = PolarsWrapper(pldf1)
        result = PolarsWrapper.concat([w1, w2], axis=1)
        assert result.get_shape() == (3, 4)

    def test_concat_axis0(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w1 = PolarsWrapper(pldf)
        w2 = PolarsWrapper(pldf.clone())
        result = PolarsWrapper.concat([w1, w2], axis=0)
        assert result.get_shape() == (6, 3)

    def test_to_array(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf)
        arr = w.to_array()
        assert isinstance(arr, np.ndarray)
        assert arr.shape == (3, 3)

    # from_output
    def test_from_output_none(self):
        from mllabs._data_wrapper import PolarsWrapper
        assert PolarsWrapper.from_output(None) is None

    def test_from_output_polars(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        result = PolarsWrapper.from_output(pldf)
        assert isinstance(result, PolarsWrapper)

    def test_from_output_ndarray_2d(self):
        from mllabs._data_wrapper import PolarsWrapper
        arr = np.array([[1, 2], [3, 4]])
        result = PolarsWrapper.from_output(arr, column_names=['a', 'b'])
        assert result.get_columns() == ['a', 'b']

    def test_from_output_ndarray_1d(self):
        from mllabs._data_wrapper import PolarsWrapper
        arr = np.array([1, 2, 3])
        result = PolarsWrapper.from_output(arr)
        assert result.get_shape() == (3, 1)

    def test_from_output_unsupported(self):
        from mllabs._data_wrapper import PolarsWrapper
        with pytest.raises(TypeError):
            PolarsWrapper.from_output("bad")

    # squeeze
    def test_squeeze_single_col(self, pldf1):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf1)
        result = w.squeeze()
        native = result.to_native()
        assert isinstance(native, pl.Series)
        assert native.to_list() == [10, 20, 30]

    def test_squeeze_multi_col(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w = PolarsWrapper(pldf)
        result = w.squeeze()
        assert isinstance(result.to_native(), pl.DataFrame)

    def test_squeeze_series(self):
        from mllabs._data_wrapper import PolarsWrapper
        s = pl.Series('y', [1, 2, 3])
        w = PolarsWrapper(s)
        result = w.squeeze()
        assert result is w

    # aggregation
    def test_mean(self, pldf):
        from mllabs._data_wrapper import PolarsWrapper
        w1 = PolarsWrapper(pldf.clone())
        w2 = PolarsWrapper(pldf.clone() * 3)
        result = PolarsWrapper.mean(iter([w1, w2]))
        native = unwrap(result)
        assert native['a'].to_list() == [2.0, 4.0, 6.0]

    def test_mode(self):
        from mllabs._data_wrapper import PolarsWrapper
        df1 = pl.DataFrame({'a': [1, 2, 3]})
        df2 = pl.DataFrame({'a': [1, 2, 4]})
        df3 = pl.DataFrame({'a': [1, 5, 4]})
        result = PolarsWrapper.mode(iter([wrap(df1), wrap(df2), wrap(df3)]))
        native = unwrap(result)
        assert native['a'].to_list()[0] == 1


# ── CudfWrapper ──

@skip_cudf
class TestCudfWrapper:
    @pytest.fixture
    def cdf(self):
        return cudf.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6], 'c': [7, 8, 9]})

    @pytest.fixture
    def cdf1(self):
        return cudf.DataFrame({'x': [10, 20, 30]})

    def test_iloc(self, cdf):
        from mllabs._data_wrapper import CudfWrapper
        w = CudfWrapper(cdf)
        result = w.iloc([0, 2])
        assert result.get_shape() == (2, 3)

    def test_select_columns(self, cdf):
        from mllabs._data_wrapper import CudfWrapper
        w = CudfWrapper(cdf)
        result = w.select_columns(['a', 'c'])
        assert result.get_columns() == ['a', 'c']

    def test_get_columns(self, cdf):
        from mllabs._data_wrapper import CudfWrapper
        w = CudfWrapper(cdf)
        assert w.get_columns() == ['a', 'b', 'c']

    def test_get_shape(self, cdf):
        from mllabs._data_wrapper import CudfWrapper
        w = CudfWrapper(cdf)
        assert w.get_shape() == (3, 3)

    def test_concat_axis1(self, cdf, cdf1):
        from mllabs._data_wrapper import CudfWrapper
        w1 = CudfWrapper(cdf)
        w2 = CudfWrapper(cdf1)
        result = CudfWrapper.concat([w1, w2], axis=1)
        assert result.get_shape() == (3, 4)

    def test_from_output_none(self):
        from mllabs._data_wrapper import CudfWrapper
        assert CudfWrapper.from_output(None) is None

    def test_squeeze_single_col(self, cdf1):
        from mllabs._data_wrapper import CudfWrapper
        w = CudfWrapper(cdf1)
        result = w.squeeze()
        assert result.to_native().ndim == 1



import pickle
import numpy as np

from ._base import Collector
from .._node_processor import resolve_columns
from .._data_wrapper import DataWrapper


class StackingCollector(Collector):
    def __init__(self, name, connector, output_var, experimenter, method='mean'):
        super().__init__(name, connector)
        self.output_var = output_var
        self.method = method
        self.columns = {}
        self._buffer = {}
        self._mem_data = {}

        self._data_cls = type(experimenter.data)
        self._index = self._build_index(experimenter)
        self._target, self._target_columns = self._build_target(experimenter)

    def _build_index(self, experimenter):
        all_valid_idx = np.concatenate([
            experimenter.valid_idx_list[i]
            for i in range(experimenter.get_n_splits())
        ])
        return experimenter.data.iloc(all_valid_idx).get_index()

    def _build_target(self, experimenter):
        target_vars = self.connector.edges.get('y') if self.connector.edges else None
        if target_vars is None:
            return None, None

        target_list = []
        target_columns = None
        temp_edges = {'_target': target_vars}
        for idx in range(experimenter.get_n_splits()):
            iterator = experimenter.get_data_valid(idx, temp_edges)
            aggregated = DataWrapper.simple(
                data_dict['_target'] for data_dict in iterator
            )
            if target_columns is None:
                target_columns = aggregated.get_columns()
            target_list.append(aggregated.to_array())

        return np.concatenate(target_list, axis=0), target_columns

    def _start(self, node):
        self._buffer[node] = []

    def _collect(self, node, idx, inner_idx, context):
        cols = resolve_columns(context['output_valid'], self.output_var)
        if len(cols) == 0:
            return
        valid_output = context['output_valid'].select_columns(cols)
        self._buffer[node].append(valid_output)
        if inner_idx == 0:
            self.columns[node] = valid_output.get_columns()

    def _end_idx(self, node, idx):
        if len(self._buffer[node]) == 0:
            return
        aggregated = self._aggregate(iter(self._buffer[node]))
        arr = aggregated.to_array()

        if self.path is not None:
            self._ensure_path()
            file_path = self.path / f"{node}.pkl"
            if file_path.exists():
                with open(file_path, 'rb') as f:
                    existing = pickle.load(f)
                existing['data'] = np.concatenate([existing['data'], arr])
            else:
                existing = {'data': arr, 'columns': self.columns[node]}
            with open(file_path, 'wb') as f:
                pickle.dump(existing, f)
        else:
            if node not in self._mem_data:
                self._mem_data[node] = {'data': arr, 'columns': self.columns[node]}
            else:
                self._mem_data[node]['data'] = np.concatenate([self._mem_data[node]['data'], arr])

        self._buffer[node] = []

    def _end(self, node):
        if node in self._buffer:
            del self._buffer[node]

    def _aggregate(self, iterator):
        if self.method == 'simple':
            return DataWrapper.simple(iterator)
        elif self.method == 'mean':
            return DataWrapper.mean(iterator)
        elif self.method == 'mode':
            return DataWrapper.mode(iterator)
        else:
            raise ValueError(f"Unsupported method: {self.method}")

    def has_node(self, node):
        if node in self._mem_data:
            return True
        if self.path is not None:
            return (self.path / f"{node}.pkl").exists()
        return False

    def reset_nodes(self, nodes):
        for node in nodes:
            if self.path is not None:
                file_path = self.path / f"{node}.pkl"
                if file_path.exists():
                    file_path.unlink()
            if node in self._mem_data:
                del self._mem_data[node]
            if node in self.columns:
                del self.columns[node]

    def save(self):
        if self.path is None:
            return
        self._ensure_path()
        config = {
            'name': self.name,
            'connector': self.connector,
            'output_var': self.output_var,
            'method': self.method,
            '_data_cls': self._data_cls,
            '_index': self._index,
            '_target': self._target,
            '_target_columns': self._target_columns,
        }
        with open(self.path / '__config.pkl', 'wb') as f:
            pickle.dump(config, f)

    @classmethod
    def load(cls, path):
        with open(path / '__config.pkl', 'rb') as f:
            config = pickle.load(f)
        obj = object.__new__(cls)
        obj.name = config['name']
        obj.connector = config['connector']
        obj.output_var = config['output_var']
        obj.method = config['method']
        obj._data_cls = config['_data_cls']
        obj._index = config['_index']
        obj._target = config['_target']
        obj._target_columns = config['_target_columns']
        obj.columns = {}
        obj._buffer = {}
        obj._mem_data = {}
        obj.path = path
        return obj

    def _load_node_data(self, node):
        if self.path is not None:
            file_path = self.path / f"{node}.pkl"
            if not file_path.exists():
                raise FileNotFoundError(f"Stacking data not found: {file_path}")
            with open(file_path, 'rb') as f:
                node_info = pickle.load(f)
            return node_info['data'], node_info['columns']
        else:
            if node not in self._mem_data:
                raise KeyError(f"Stacking data not found in memory: {node}")
            return self._mem_data[node]['data'], self._mem_data[node]['columns']

    def _get_saved_nodes(self):
        if self.path is not None:
            if not self.path.exists():
                return []
            return [f.stem for f in self.path.glob("*.pkl") if not f.stem.startswith('__')]
        else:
            return list(self._mem_data.keys())

    def get_dataset(self, nodes=None, include_target=True):
        node_names = self._get_nodes(nodes, self._get_saved_nodes())

        node_data = []
        column_names = []
        for i in node_names:
            data, columns = self._load_node_data(i)
            node_data.append(data)
            column_names.extend(columns)

        all_data = np.concatenate(node_data, axis=1)
        all_columns = list(column_names)

        wrapped_data = self._data_cls.from_output(all_data, all_columns, self._index)
        if include_target and self._target is not None:
            wrapped_data = self._data_cls.concat([
                wrapped_data, 
                self._data_cls.from_output(self._target, self._target_columns, self._index)
            ], axis=1)

        return wrapped_data.to_native()



import os
import shutil
import pickle as pkl
import traceback
import warnings
import numpy as np
from pathlib import Path

from ._data_wrapper import wrap, unwrap
from ._trainobj import TrainStageObj, TrainHeadObj
from ._node_processor import resolve_columns


class Trainer:

    def __init__(self, name, pipeline, data, path, splitter, splitter_params, cache, logger):
        self.name = name
        self.pipeline = pipeline
        self.data = data
        self.path = Path(path)
        self.splitter = splitter
        self.splitter_params = splitter_params
        self.cache = cache
        self.logger = logger

        self.selected_stages = []
        self.selected_heads = []

        self.node_objs = {}

        self.split_indices = self._make_splits()
        self.save()

    # ------------------------------------------------------------------
    # node selection
    # ------------------------------------------------------------------

    def select_head(self, nodes):
        node_names = self.pipeline.get_node_names(nodes)

        selected = set(self.selected_stages + self.selected_heads)
        for name in node_names:
            selected.add(name)
            self._collect_upstream(name, selected)

        all_ordered = self.pipeline._get_affected_nodes([None])
        self.selected_stages = []
        self.selected_heads = []
        for n in all_ordered:
            if n is None or n not in selected:
                continue
            grp = self.pipeline.get_grp(self.pipeline.get_node(n).grp)
            if grp.role == 'stage':
                self.selected_stages.append(n)
            elif grp.role == 'head':
                self.selected_heads.append(n)

    def _collect_upstream(self, node_name, selected):
        node_attrs = self.pipeline.get_node_attrs(node_name)
        for key, edge_list in node_attrs['edges'].items():
            for source_node, var in edge_list:
                if source_node is not None and source_node not in selected:
                    selected.add(source_node)
                    self._collect_upstream(source_node, selected)

    def reset_nodes(self, nodes):
        selected_set = set(self.selected_stages + self.selected_heads)
        affected = set(n for n in nodes if n in selected_set)

        queue = list(affected)
        while queue:
            n = queue.pop(0)
            node = self.pipeline.get_node(n)
            for downstream in node.output_edges:
                if downstream in selected_set and downstream not in affected:
                    affected.add(downstream)
                    queue.append(downstream)

        for node_name in affected:
            if node_name in self.node_objs:
                node_obj = self.node_objs.pop(node_name)
                if os.path.isdir(node_obj.path):
                    shutil.rmtree(node_obj.path)

        if self.cache is not None:
            self.cache.clear_nodes(affected)

        self.save()

    # ------------------------------------------------------------------
    # train
    # ------------------------------------------------------------------

    def train(self):
        selected_set = set(self.selected_stages + self.selected_heads)
        stage_set = set(self.selected_stages)

        target_nodes = []
        for name in self.pipeline._get_affected_nodes([None]):
            if name not in selected_set:
                continue
            if name in self.node_objs:
                continue
            target_nodes.append(name)

        for node_name in target_nodes:
            if node_name in stage_set:
                node_obj = TrainStageObj(self._get_node_path(node_name))
            else:
                node_obj = TrainHeadObj(self._get_node_path(node_name))
            node_obj.start_build()
            self.node_objs[node_name] = node_obj

        n_splits = self.get_n_splits()
        self.logger.start_progress("Train", len(target_nodes))
        for ni, node_name in enumerate(target_nodes):
            self.logger.update_progress(ni)
            self.logger._progress[-1][0] = node_name
            node_obj = self.node_objs[node_name]
            node_attrs = self.pipeline.get_node_attrs(node_name)
            self.logger.start_progress("Split", n_splits)
            for split_idx, data_dict in enumerate(self.get_data(node_attrs['edges'])):
                self.logger.update_progress(split_idx)
                if node_obj.status == 'error':
                    break
                try:
                    with warnings.catch_warnings(record=True) as caught:
                        warnings.simplefilter("always")
                        node_obj.build_split(split_idx, node_attrs, data_dict, self.logger)
                        for w in caught:
                            self.logger.warning(f"[{node_name}] split {split_idx}: {w.category.__name__}: {w.message}")
                except Exception as e:
                    node_obj.status = 'error'
                    node_obj.error = {
                        'type': type(e).__name__,
                        'message': str(e),
                        'traceback': traceback.format_exc(),
                        'split': split_idx,
                    }
                    self.logger.info(f"[{node_name}] Train error at split {split_idx}: {type(e).__name__}: {e}")
                    self.logger.info(traceback.format_exc())
                    self.cache.clear_nodes([node_name])
            self.logger.end_progress(n_splits)
            if node_obj.status != 'error':
                node_obj.end_build()
        self.logger.end_progress(len(target_nodes))

        error_nodes = [n for n in target_nodes if self.node_objs[n].status == 'error']
        if error_nodes:
            self.logger.info(f"Train complete: {len(target_nodes) - len(error_nodes)}/{len(target_nodes)} node(s), {len(error_nodes)} error(s): {error_nodes}")
        else:
            self.logger.info(f"Train complete: {len(target_nodes)} node(s)")

        self.save()

    # ------------------------------------------------------------------
    # data resolution
    # ------------------------------------------------------------------

    def _make_splits(self):
        if self.splitter is None:
            return None
        data_native = unwrap(self.data)
        split_params = {'X': data_native}
        for k, v in self.splitter_params.items():
            split_params[k] = unwrap(self.data.select_columns(v))
        return [
            (train_idx, valid_idx)
            for train_idx, valid_idx in self.splitter.split(**split_params)
        ]

    def get_n_splits(self):
        if self.split_indices is None:
            return 1
        return len(self.split_indices)

    def get_data(self, edges):
        data_dict = {}
        for key, edge_list in edges.items():
            key_data_list = []
            for node_name, var in edge_list:
                key_data_list.append(self.get_node_output(node_name, var))
            data_dict[key] = zip(*key_data_list)

        for _ in range(self.get_n_splits()):
            result = {}
            for k, it in data_dict.items():
                z = next(it)
                train_sub, valid_sub = [], []
                for train_data, valid_data in z:
                    train_sub.append(train_data)
                    if valid_data is not None:
                        valid_sub.append(valid_data)
                train_concat = type(train_sub[0]).concat(train_sub, axis=1)
                if valid_sub:
                    valid_concat = type(valid_sub[0]).concat(valid_sub, axis=1)
                else:
                    valid_concat = None
                result[k] = train_concat, valid_concat
            yield result

    def get_node_output(self, node, v=None):
        if node is None:
            if self.split_indices is None:
                data = self.data if v is None else self.data.select_columns(v)
                yield data, None
            else:
                for train_idx, valid_idx in self.split_indices:
                    train_data = self.data.iloc(train_idx)
                    valid_data = self.data.iloc(valid_idx)
                    if v is not None:
                        train_data = train_data.select_columns(v)
                        valid_data = valid_data.select_columns(v)
                    yield train_data, valid_data
            return

        cached = self.cache.get_data(node, "train_all", 0)
        if cached is not None:
            for (train_result, valid_result), (obj, _, _) in zip(cached, self.node_objs[node].get_obj()):
                if v is not None:
                    X = resolve_columns(train_result, v, processor=obj)
                    train_result = train_result.select_columns(X)
                    if valid_result is not None:
                        valid_result = valid_result.select_columns(X)
                yield train_result, valid_result
            return

        node_attrs = self.pipeline.get_node_attrs(node)
        cache_data = []
        for data_dict, (obj, train_, info) in zip(self.get_data(node_attrs['edges']), self.node_objs[node].get_obj()):
            train_X, valid_X = data_dict['X']

            if train_ is None:
                train_result = obj.process(train_X)
            else:
                train_result = train_

            valid_result = obj.process(valid_X) if valid_X is not None else None

            cache_data.append((train_result, valid_result))

            if v is not None:
                X = resolve_columns(train_result, v, processor=obj)
                train_result = train_result.select_columns(X)
                if valid_result is not None:
                    valid_result = valid_result.select_columns(X)

            yield train_result, valid_result
        self.cache.put_data(node, "train_all", 0, cache_data)

    def get_node_data(self, node):
        node_attrs = self.pipeline.get_node_attrs(node)
        return self.get_data(node_attrs['edges'])
    
    def process(self, data, v=None):
        data = wrap(data)
        ordered = [
            name for name in self.pipeline._get_affected_nodes([None])
            if name in set(self.selected_stages + self.selected_heads)
        ]
        stage_set = set(self.selected_stages)

        obj_iters = {name: self.node_objs[name].get_obj() for name in ordered}

        for _ in range(self.get_n_splits()):
            data_dicts = {None: (data, None)}
            head_outputs = []

            for name in ordered:
                obj, _, _ = next(obj_iters[name])
                node_attrs = self.pipeline.get_node_attrs(name)
                output = self._process_node(obj, data_dicts, node_attrs['edges'])

                if name in stage_set:
                    data_dicts[name] = (output, obj)
                else:
                    if v is not None:
                        cols = resolve_columns(output, v, processor=obj)
                        output = output.select_columns(cols)
                    head_outputs.append(output)

            if len(head_outputs) == 1:
                yield head_outputs[0]
            else:
                yield type(head_outputs[0]).concat(head_outputs, axis=1)

    def _process_node(self, obj, data_dicts, edges):
        input_data = self._get_process_data(data_dicts, edges)
        return obj.process(input_data['X'])

    def _get_process_data(self, data_dicts, edges):
        result = {}
        for key, edge_list in edges.items():
            parts = []
            for src_node, var in edge_list:
                src, obj = data_dicts[src_node]
                if var is not None:
                    cols = resolve_columns(src, var, processor=obj)
                    src = src.select_columns(cols)
                parts.append(src)
            if len(parts) == 1:
                result[key] = parts[0]
            else:
                result[key] = type(parts[0]).concat(parts, axis=1)
        return result

    # ------------------------------------------------------------------
    # to_inferencer
    # ------------------------------------------------------------------

    def to_inferencer(self, v=None):
        from ._inferencer import Inferencer

        all_selected = self.selected_stages + self.selected_heads
        for name in all_selected:
            obj = self.node_objs.get(name)
            if obj is None or obj.status != 'built':
                raise RuntimeError(f"Node '{name}' is not built. Run train() first.")

        pipeline = self.pipeline.copy_nodes(self.selected_heads)
        node_objs = {
            name: [obj for obj, _, _ in self.node_objs[name].get_obj()]
            for name in all_selected
        }
        return Inferencer(pipeline, list(self.selected_stages), list(self.selected_heads),
                          self.get_n_splits(), node_objs, v=v)

    # ------------------------------------------------------------------
    # path helpers
    # ------------------------------------------------------------------

    def _get_grp_path(self, grp):
        if isinstance(grp, str):
            grp = self.pipeline.get_grp(grp)
        path_parts = [grp.name]
        current = self.pipeline.get_grp(grp.parent)
        while current is not None:
            path_parts.insert(0, current.name)
            current = self.pipeline.get_grp(current.parent)
        return self.path / '/'.join(path_parts)

    def _get_node_path(self, node_name):
        node = self.pipeline.get_node(node_name)
        return self._get_grp_path(node.grp) / node.name

    # ------------------------------------------------------------------
    # save / load
    # ------------------------------------------------------------------

    def save(self):
        if self.path is None:
            return
        self.path.mkdir(parents=True, exist_ok=True)
        save_data = {
            'name': self.name,
            'splitter': self.splitter,
            'splitter_params': self.splitter_params,
            'selected_stages': self.selected_stages,
            'selected_heads': self.selected_heads,
            'node_obj_keys': list(self.node_objs.keys()),
            'split_indices': self.split_indices,
        }
        with open(self.path / '__trainer.pkl', 'wb') as f:
            pkl.dump(save_data, f)

    @classmethod
    def _load(cls, path, pipeline, data, cache, logger):
        path = Path(path)
        with open(path / '__trainer.pkl', 'rb') as f:
            save_data = pkl.load(f)

        trainer = object.__new__(cls)
        trainer.name = save_data['name']
        trainer.pipeline = pipeline
        trainer.data = data
        trainer.path = path
        trainer.splitter = save_data['splitter']
        trainer.splitter_params = save_data['splitter_params']
        trainer.cache = cache
        trainer.logger = logger
        trainer.selected_stages = save_data['selected_stages']
        trainer.selected_heads = save_data['selected_heads']
        trainer.split_indices = save_data['split_indices']

        stage_set = set(trainer.selected_stages)
        trainer.node_objs = {}
        for node_name in save_data['node_obj_keys']:
            node_path = trainer._get_node_path(node_name)
            if os.path.isdir(node_path):
                if node_name in stage_set:
                    node_obj = TrainStageObj(node_path)
                else:
                    node_obj = TrainHeadObj(node_path)
                node_obj.load()
                trainer.node_objs[node_name] = node_obj

        return trainer
